/*
+---------------------------------------------------------------------------------------+
|																						|
|	Author: Jonas Eberhard | www.joseng.de												|
|	Datei:	twi.c																		|
|	Beschreibung: I2C (TWI) Libary														|
|																						|
|																						|
|	Prozessor:																			|
|	FuseBits: 																			|
|																						|
+---------------------------------------------------------------------------------------+

+---------------------------------------------------------------------------------------+
|																						|
|	Libary getestet mit folgenden Bausteinen:											|
|		Stand Verffentlichung in diesem Projekt, Liste wird laufend erweitert			|
|																						|
|	PCF8574 (IO Expander), TDA8425 (Audio COntroller)									|
|																						|
+---------------------------------------------------------------------------------------+
*/


//#include <util/twi.h>		//TWI Bitmask Definitions 
#include <avr/io.h>
#include <stdint.h>

#include "twi.h"
#include <util/delay.h>  


//I2C Init, Baurate wird gesetzt, Angabe in Hertz
uint8_t twi_init(uint16_t baudrate)
{
	uint8_t dummy = 0;
	
	//Bitrate setzten
	TWSR = 0; 		//Prescaler = 0
	dummy = ((F_CPU/baudrate)/2)-8;
	if(dummy > 10)
	{
		TWBR = dummy;
		return 1;
	}
	return 0;
}


//Startbedingung auf I2C senden und danach Slave Baustein adressieren mit R/W
//Slave_adress = I2C Bausteinadresse, zum lesen von I2C Slave muss noch eine 1 zur Adresse addiert werden, dafr stehen
//die defines TWI_WRITE und TWI_READ zur Verfgung, Schreiben: Adresse+TWI_WRITE (im Grunde sinnlos), Lesen: Adresse+TWI_READ
uint8_t twi_start(uint8_t slave_adress)
{
	//I2C Startbedingung senden auslsen
	TWCR = (1<<TWEN) | (1<< TWINT) | (1<<TWSTA);
	//Warten bis Interrupt Bit gesetzt ist -> Aktion fertig ausgefhrt
	while (!(TWCR & (1<<TWINT)));
	//Wenn nicht erfolgreich gesendet dann raus mit Code 0 = nicht erfolgreich
	if((TWI_STATUS != TWI_START) && (TWI_STATUS != TWI_REP_START))
	{
		return 0;
	}
	
	//Adresse und Richtung in Datenregister schreiben
	TWDR = slave_adress;
	//TWINT Bit lschen -> Daten werden bertragen
	TWCR = (1<<TWINT) | (1<<TWEN);
	//Warten bis Interrupt Bit gesetzt ist -> Aktion fertig ausgefhrt
	while (!(TWCR & (1<<TWINT)));
	//Wenn nicht erfolgreich gesendet dann raus mit Code 0 = nicht erfolgreich
	if((TWI_STATUS != TWI_MTX_ADR_ACK) && (TWI_STATUS != TWI_MRX_ADR_ACK))
	{	
		return 0;
	}
		
	return 1;	//Wenn bis hier hin alles glatt gelaufen ist, dann raus mit Code 1 = alles ok
}


//Diese Funktion sendet ein Byte auf dem I2C Bus und wartet bis es gesendet wurde
uint8_t twi_send_byte(uint8_t value)
{
	//Datenregister nur beschreiben wenn TWINT 1 ist
	if((TWCR & 0b10000000))
	{
		//Datenregister beschreiben
		TWDR = value;
		//TWINT Bit lschen -> Daten werden bertragen
		TWCR = (1<<TWINT) | (1<<TWEN);
		//Warten bis Interrupt Bit gesetzt ist -> Aktion fertig ausgefhrt
		while (!(TWCR & (1<<TWINT)));
		//Wenn nicht erfolgreich gesendet dann raus mit Code 0 = nicht erfolgreich
		if((TWI_STATUS != TWI_MTX_DATA_ACK))
		{
			return 0;
		}
		//Wenn alles gut mit 1 raus -> erfolgreich
		return 1;
	}
	
	return 0;
}


//Diese Funktion sendet mehrere Bytes hintereinander auf den I2C Bus
//len = Anzahl Bytes zu bertragen, buffer = array mit Bytes zum senden
uint8_t twi_send_bytes(uint8_t len, uint8_t *buffer)
{
	uint8_t i;
	
	for(i=0; i<len; i++)
	{
		if(!twi_send_byte(buffer[i]))
			return 0;
	}
	
	return 1;
}


//Daten von Slave lesen, nachdem dieser mit read adressiert wurde, ACK senden
uint8_t twi_read_byte_ack(void)
{
	//TWINT Bit lschen -> Daten werden bertragen, TWEN Bit setzten, dass bei erfolgreicher bertragung ein ACK zum Slave gesendet wird
	TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWEA);
	//Warten bis Interrupt Bit gesetzt ist -> Aktion fertig ausgefhrt
	while (!(TWCR & (1<<TWINT)));
	
	return TWDR;
}


//Daten von Slave lesen, nachdem dieser mit read adressiert wurde, kein ACK senden -> damit wurde hier das letzte Byte gelesen
uint8_t twi_read_byte_nack(void)
{
	//TWINT Bit lschen -> Daten werden bertragen, TWEN Bit setzten, dass bei erfolgreicher bertragung ein ACK zum Slave gesendet wird
	TWCR = (1<<TWINT) | (1<<TWEN);
	//Warten bis Interrupt Bit gesetzt ist -> Aktion fertig ausgefhrt
	while (!(TWCR & (1<<TWINT)));
	
	return TWDR;	
}


//Mehrere Bytes hintereinander vom I2C Slave lesen
uint8_t twi_read_bytes(uint8_t len, uint8_t *buffer)
{
	uint8_t i;
	
	for(i=0; i<len-1; i++)
	{
		buffer[i] = twi_read_byte_ack();
	}
	buffer[i+1] = twi_read_byte_nack();
	
	return 1;
}


//I2C Stopbedingung senden und damit Kommunikation abschlieen
uint8_t twi_stop(void)
{
	//Stopbedingung auslsen
	TWCR = (1<<TWINT) | (1<<TWEN) | (1<<TWSTO);
	//while ((TWCR & (1<<TWSTO)));
	return 1;
}


void blink(uint8_t anz)
{
	uint8_t i;
	PORTB &= ~(1<<0);
	_delay_ms(50);
	for(i=0; i<anz; i++)
	{
		PORTB |= (1<<0);
		_delay_ms(100);
		PORTB &= ~(1<<0);
		_delay_ms(100);		
	}
	PORTB &= ~(1<<0);
}